#include "PortListener.h"
#include <QtDebug>

PortListener::PortListener(const QString &portName)
{
    qDebug() << "hi there";
    this->port = new QextSerialPort(portName, QextSerialPort::EventDriven);
    port->setBaudRate(BAUD115200);
    //port->setBaudRate(BAUD9600);
    port->setFlowControl(FLOW_OFF);
    port->setParity(PAR_NONE);
    port->setDataBits(DATA_8);
    port->setStopBits(STOP_2);

    if (port->open(QIODevice::ReadWrite) == true) {
        connect(port, SIGNAL(readyRead()), this, SLOT(onReadyRead()));
        connect(port, SIGNAL(dsrChanged(bool)), this, SLOT(onDsrChanged(bool)));
        if (!(port->lineStatus() & LS_DSR)){
            qDebug() << "warning: device is not turned on";
            bThereIsConnection = false;
        }else
            bThereIsConnection = true;

        qDebug() << "listening for data on" << port->portName();
        //bThereIsConnection = true;
    }
    else {
        qDebug() << "device failed to open:" << port->errorString();
        bThereIsConnection = false;
    }
}
PortListener::PortListener(const QString &portName,int nBaudRate, int nParity, int nBits, int nStopBits, int nFlow){
    vSetPortSettings(portName,nBaudRate,nParity,nBits,nStopBits,nFlow);
}

PortListener::~PortListener(){
    delete this->port;
}

void PortListener::onReadyRead()
{
    QByteArray bytes;
    int a = port->bytesAvailable();
    bytes.resize(a);
    port->read(bytes.data(), bytes.size());
//    qDebug() << "bytes read:" << bytes.size();
//    qDebug() << "bytes:" << bytes;

    emit vBytesRead(bytes);
}

void PortListener::onDsrChanged(bool status)
{
    if (status)
        qDebug() << "device was turned on";
    else
        qDebug() << "device was turned off";
}

void PortListener::vSetPortSettings(const QString &portName,int nBaudRate,int nParity,int nBits,int nStopBits,int nFlow){
    //qDebug() << "hi there";
    this->port = new QextSerialPort(portName, QextSerialPort::EventDriven);
//    port->setBaudRate(BAUD56000);
//    port->setBaudRate(BAUD115200);
    port->setBaudRate((BaudRateType)nBaudRate);
//    port->setFlowControl(FLOW_OFF);
    port->setFlowControl((FlowType)nFlow);
//    port->setParity(PAR_NONE);
    port->setParity((ParityType)nParity);
//    port->setDataBits(DATA_8);
    port->setDataBits((DataBitsType)nBits);
//    port->setStopBits(STOP_2);
    port->setStopBits((StopBitsType)nStopBits);

    if (port->open(QIODevice::ReadWrite) == true) {
        connect(port, SIGNAL(readyRead()), this, SLOT(onReadyRead()));
        connect(port, SIGNAL(dsrChanged(bool)), this, SLOT(onDsrChanged(bool)));
        if (!(port->lineStatus() & LS_DSR))
            qDebug() << "warning: device is not turned on";
        qDebug() << "listening for data on" << port->portName();
        bThereIsConnection = true;
    }
    else {
        qDebug() << "device failed to open:" << port->errorString();
        bThereIsConnection = false;
    }
}

bool PortListener::bWrite(QByteArray qba){
    if(bThereIsConnection){
        //qbaToSend = qba;
        port->write(qba);
        //qDebug() << qba;
        //QTimer::singleShot(100, this, SLOT(vWrite()));
    }
    return bThereIsConnection;
}

bool PortListener::bCheckRTS(){
    return port->lineStatus() & LS_RTS;
}

bool PortListener::bCheckDTR(){
    return port->lineStatus() & LS_DTR;
}

bool PortListener::bCheckDSR(){
    return port->lineStatus() & LS_DSR;
}

bool PortListener::bCheckDCD(){
    return port->lineStatus() & LS_DCD;
}

void PortListener::vSetRTS(bool b){
    port->setRts(b);
}

void PortListener::vSetDTR(bool b){
    port->setDtr(b);
}

QString PortListener::qszLineStatus(){
  ulong n = port->lineStatus();
  QString s = "";
  if (n & LS_CTS){ if (!s.isEmpty()) s += ","; s += " CTS";}
  if (n & LS_DSR){ if (!s.isEmpty()) s += ","; s += " DSR";}
  if (n & LS_DCD){ if (!s.isEmpty()) s += ","; s += " DCD";}
  if (n & LS_RI){ if (!s.isEmpty()) s += ","; s += " RI";}
  if(s.isEmpty())
      s = "Nothing detected";
  return s;
}
