#include "qtusbtwi.h"

#include <QDebug>

qtUsbTwi::qtUsbTwi(PortListener *l, QObject *parent) :
    QObject(parent)
{
    listener = l;
    listLastDetectedTwiDevices = listDetectedTwiDevices();
}

void qtUsbTwi::vWait1us(){
    //QTest::qWait(1);
    usleep(K_mS);
}

void qtUsbTwi::vSCL(bool b){
    listener->vSetRTS(b);
}

void qtUsbTwi::vSDA(bool b){
    listener->vSetDTR(!b);
}

bool qtUsbTwi::bCheckSCL(){
    return listener->bCheckRTS();
}

bool qtUsbTwi::bCheckSDAo(){
    return !listener->bCheckDTR();
}

bool qtUsbTwi::bCheckSDAi(){
    //return listener->bCheckDCD();
    return listener->bCheckDSR();
}

void qtUsbTwi::vTwiStart(){
/* A start condition looks like this:

     SDA H ------\             The chip issuing the Start condition first pulls
         L        \-------     the SDA (data) line low. And next
                               pulls the SCL (clock) line low.
     SCL H ---------\
         L           \----
*/
        vSDA(true);
        vSCL(true);
        vWait1us();
        vSDA(false);
        vWait1us();
        vSCL(false);
}

void qtUsbTwi::vTwiStop(){
/* A stop condition looks like this:

     SDA H           /----     The Bus master first releases the SCL and then
         L ---------/          the SDA line

     SCL H        /-------
         L ------/
*/
        vSCL(false); // should already be the case
        vSDA(false);
        vSCL(true);
        vWait1us();
        vSDA(true);
        vWait1us();
}

void qtUsbTwi::vTwiTxBit(bool b){
    vSCL(false);
    vWait1us();
    if (b)
        vSDA(true);
    else
        vSDA(false);
    vSCL(true);
    vWait1us();
    vSCL(false);
}

bool qtUsbTwi::bTwiRxBit(){
        bool b;

        vSCL(false);
        vWait1us();
        vSCL(true);
        vWait1us();
        b=bCheckSDAi();
        vSCL(false);

        return b;
}

bool qtUsbTwi::bTwiRxAck(){
    bool bAck = false;
    int i=0;

    vSDA(true);
    vWait1us();
    vSCL(true);
    while(i<10 && bAck==false){
        i++;
        if (! bCheckSDAi()){
            // sda is low -> ack found
            bAck=true;
            vWait1us();
        }else
            usleep(100*K_mS);
    }
    vSCL(true);
    return bAck;
}

void qtUsbTwi::vTwiTxAck(bool bAck){
    vTwiTxBit(!bAck);
}

bool qtUsbTwi::bTwiTxByte(quint8 byte){
    for(int i=7;i>=0;i--)
        vTwiTxBit(byte & (1<<i));
    return bTwiRxAck();
}

// if gen_ack =1 always generate ack, gen_ack =2 only generate ack if byte != '\0'
quint8 qtUsbTwi::byTwiRxByte(int gen_ack){
    quint8 data=0;

    vSDA(true);
    for(int i=7;i>=0;i--)
        if (bTwiRxBit())//{
            data |= 1<<i;
//            qDebug() << "1";
//        }else{
//            qDebug() << "0";
//        }
    if (gen_ack==2)
        if (data == '\0')
            gen_ack=0;
    vTwiTxAck(gen_ack);

    return data;
}

bool qtUsbTwi::bTwiTxString(char* sz){ // string no longer than MAX_BUF_LEN
        int i=0;
        bool bAck = true;

        while(i<MAX_BUF_LEN && *(sz+i)){
                // we expect ack for all bytes:
                if (!bTwiTxByte((quint8) *(sz+i))){
                        bAck=false;
                }
                i++;
        }
        // the last byte (zeor byte) may get a nack:
        bTwiTxByte(0);
        usleep(100*K_mS); // give the uC a chance to poll the data
        return bAck;
}

void qtUsbTwi::vTwiRxString(char *szRcv){
        int i=0;
        *szRcv=1;
        while(i<MAX_BUF_LEN){
                *(szRcv+i)=byTwiRxByte(2);
                if (*(szRcv+i)=='\0'){
                        break;
                }
                i++;
        }
        // in case we did not get a zero:
        *(szRcv+MAX_BUF_LEN)='\0';
}

bool qtUsbTwi::bWrite8591(quint8 device, quint8 control, quint8 data){
    bool bWriteCorrect = true;

    vTwiStart();
    bWriteCorrect &= bTwiTxByte(device);
    bWriteCorrect &= bTwiTxByte(control);
    bWriteCorrect &= bTwiTxByte(data);
    vTwiStop();

    return bWriteCorrect;
}

quint8 qtUsbTwi::bRead8591(quint8 device, quint8 *pByVal, quint8 byQuants){
    bool bWriteCorrect = true;
//    quint8 i;

    //byQuants --;
    for(int j = 0 ; j < byQuants ; j++){
        bWrite8591(device, j, 0x00);
//        vWait1us();
        vTwiStart();
        bWriteCorrect &= bTwiTxByte(device | 0x01);
//        for(i = 0; i < 2; i++){
//            pByVal[j] = byTwiRxByte(1);
//            vWait1us();
//        }
        pByVal[j] = byTwiRxByte(0);
        vTwiStop();
//        vWait1us();
    }

    return bWriteCorrect;
}

bool qtUsbTwi::bWriteLm75B(quint8 device, quint8 control, quint8 data){
    bool bWriteCorrect = true;

    vTwiStart();
    bWriteCorrect &= bTwiTxByte(device);
    bWriteCorrect &= bTwiTxByte(control);
    bWriteCorrect &= bTwiTxByte(data);
    vTwiStop();

    return bWriteCorrect;
}

bool qtUsbTwi::bWriteLm75B(quint8 device, quint8 pointer){
    bool bWriteCorrect = true;

    vTwiStart();
    bWriteCorrect &= bTwiTxByte(device);
    bWriteCorrect &= bTwiTxByte(pointer);
    vTwiStop();

    return bWriteCorrect;
}

bool qtUsbTwi::bReadLm75B(quint8 device, quint16 *pByVal){
    bool bWriteCorrect = true;

    bWriteLm75B(device, 0x00);
    vTwiStart();
    bWriteCorrect &= bTwiTxByte(device | 0x01);
    *pByVal = ((quint16)byTwiRxByte(1)) << 8;
    *pByVal |= ((quint16)byTwiRxByte(0));
    vTwiStop();

    *pByVal >>= 5;

    return bWriteCorrect;
}

bool qtUsbTwi::bWriteTwi(quint8 device,quint8 data){
    bool bWriteCorrect = true;

    vTwiStart();
    bWriteCorrect &= bTwiTxByte(device);
    bWriteCorrect &= bTwiTxByte(data);
    vTwiStop();

    return bWriteCorrect;
}

bool qtUsbTwi::bReadTwi(quint8 device,quint8* data){
    bool bWriteCorrect = true;

//    vTwiStart();
//    bWriteCorrect &= bTwiTxByte(device);
//    bWriteTwi(device,0xFF); ////////////////////////////////////////
    bWriteTwi(device,*data); // Data should come with pull-ups + outputs to 1 (i.e.: To read all byte 0xFF)
    vTwiStart();
    bWriteCorrect &= bTwiTxByte(device | 0x01);
    *data = byTwiRxByte(0);
    vTwiStop();
//    QString qsz = trUtf8("0x%1").arg(QString::number(*data,16));
//    qDebug() << qsz.toLatin1();
    return bWriteCorrect;
}

QList<QString> qtUsbTwi::listDetectedTwiDevices(){
    QList<QString> qlAdd;

    quint8 byValor = 0x00;
    QString qsz;

    qDebug() << "Trying to write 0x00 to all address:";
    qlAdd.clear();
    for(int n= 0x00 ; n < 0x100 ; n+=2){
        if(!bWriteTwi(n,byValor)){
            qsz = QString::number((quint8)n>>1,16);
            qsz = qsz.toUpper();
            qlAdd.append(qsz);
            qsz = trUtf8("0x%1").arg(qsz);
            qDebug() << qsz.toLatin1();
            //printf ("0x%X ",n>>1);
        }//else
         //   qDebug() << "- ";
    }
    listLastDetectedTwiDevices = qlAdd;
    return qlAdd;
}
